﻿using System;
using System.Threading;
using System.IO;

using AppModule.InterProcessComm;
using AppModule.NamedPipes;

namespace WinCommand.CommandPipe
{

    public sealed class ServerNamedPipe : IDisposable
    {
        internal Thread PipeThread;
        internal ServerPipeConnection PipeConnection;
        internal bool Listen = true;
        internal DateTime LastAction;
        private bool disposed = false;
        private PipeManager _pipeMgr;

        public delegate void PipeEventHandler(object sender, PipeEventArgs e);
        public event PipeEventHandler PipeEvent;

        internal ServerNamedPipe(string name, uint outBuffer, uint inBuffer, int maxReadBytes, bool secure, PipeManager pipeMgr)
        {
            PipeConnection = new ServerPipeConnection(name, outBuffer, inBuffer, maxReadBytes, secure);
            PipeThread = new Thread(new ThreadStart(PipeListener));
            PipeThread.IsBackground = true;
            PipeThread.Name = "Pipe Thread " + this.PipeConnection.NativeHandle.ToString();
            LastAction = DateTime.Now;
            _pipeMgr = pipeMgr;
        }

        private void OnPipeEvent(PipeEventArgs e)
        {
            if (PipeEvent != null)
                PipeEvent(this, e);
        }

        private void PipeListener()
        {
            CheckIfDisposed();
            try
            {
                Listen = _pipeMgr.Listen;
                OnPipeEvent(new PipeEventArgs("Pipe " + this.PipeConnection.NativeHandle.ToString() + ": new pipe started", PipeEventArgs.PipeEvent.PipeStarted));
                //DM 1
                while (Listen)
                {
                    LastAction = DateTime.Now;
                    string request = PipeConnection.Read();
                    LastAction = DateTime.Now;
                    if (request.Trim() != "")
                    {
                        PipeConnection.Write(_pipeMgr.HandleRequest(request));
                        //DM 2
                        OnPipeEvent(new PipeEventArgs("Pipe " + this.PipeConnection.NativeHandle.ToString() + ": request handled", PipeEventArgs.PipeEvent.RequestHandled));
                    }
                    else
                    {
                        PipeConnection.Write("Error: bad request");
                    }
                    LastAction = DateTime.Now;
                    PipeConnection.Disconnect();
                    if (Listen)
                    {
                        //DM 3
                        OnPipeEvent(new PipeEventArgs("Pipe " + this.PipeConnection.NativeHandle.ToString() + ": listening", PipeEventArgs.PipeEvent.Listening));
                        Connect();
                    }
                    _pipeMgr.WakeUp();
                }
            }
            catch (System.Threading.ThreadAbortException ex) { }
            catch (System.Threading.ThreadStateException ex) { }
            catch (Exception ex)
            {
                // Log exception
            }
            finally
            {
                this.Close();
            }
        }
        internal void Connect()
        {
            CheckIfDisposed();
            PipeConnection.Connect();
        }
        internal void Close()
        {
            CheckIfDisposed();
            this.Listen = false;
            _pipeMgr.RemoveServerChannel(this.PipeConnection.NativeHandle);
            this.Dispose();
        }
        internal void Start()
        {
            CheckIfDisposed();
            PipeThread.Start();
        }
        private void CheckIfDisposed()
        {
            if (this.disposed)
            {
                throw new ObjectDisposedException("ServerNamedPipe");
            }
        }
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }
        private void Dispose(bool disposing)
        {
            if (!this.disposed)
            {
                PipeConnection.Dispose();
                if (PipeThread != null)
                {
                    try
                    {
                        PipeThread.Abort();
                    }
                    catch (System.Threading.ThreadAbortException ex) { }
                    catch (System.Threading.ThreadStateException ex) { }
                    catch (Exception ex)
                    {
                        // Log exception
                    }
                }
            }
            disposed = true;
        }
        ~ServerNamedPipe()
        {
            Dispose(false);
        }
    }
}